
Acts = new Mongo.Collection("acts");
ActExecutions = new Mongo.Collection("actExecutions");


if (Meteor.isServer) {
    Meteor.publish("acts", function () {

        // for debugging of loading spinner
        if( process.env.NODE_ENV === "development" )
            Meteor._sleepForMs(500);

        var self = this;
        if( self.userId )
            return Acts.find({ownerUserId: self.userId});
        else
            return [];
    });

    // necessary for users who are not signed and are playing with the demo
    Meteor.publish("actById", function (actId) {
        var self = this;
        return Acts.find({_id: actId, $or: [ { ownerUserId: self.userId}, { ownerUserId: null } ]});
    });

    Meteor.publish("actExecutionsById", function (actExecutionId) {
        return ActExecutions.find(actExecutionId);
    });
}


// a helper function to determine whether some record field was modified, check its new value,
// update existing record with it and update appropriate MongoDB insert/update modifier
var updateAndCheckFields = function(fieldSpecs, newRecord, modifier, existingRecord) {
    var updated = false;
    _.each( fieldSpecs, function( fieldSpec, fieldName ) {
        var newVal = newRecord[fieldName];

        // if updating existing record and a field is undefined or unchanged, its value won't be updated
        if( existingRecord && (newVal === undefined || newVal === existingRecord[fieldName]) )
            return;

        // all other checks are covered in generated checkPattern
        if( !Match.test(newVal, fieldSpec.checkPattern) )
            throw new Meteor.Error("act-check-failed", fieldName);

        // save new value
        if( existingRecord )
            existingRecord[fieldName] = newVal;
        modifier[fieldName] = newVal;
        updated = true;
    });
    return updated;
};


Meteor.methods({

    // inserts a new act to database and returns it (the return value might be different than original 'act'!)
    'insertAct': function(act) {
        var self = this;
        if( self.userId !== null )
            check(self.userId, String);
        check(act, Match.ObjectIncluding({
            type: String
        }));
        check( act._id, undefined );
        if( !actFieldsForType[act.type] )
            throw new Meteor.Error("act-type-invalid");

        // NOTE: we don't insert 'act' directly but rather copy known fields into a new object 'modifier',
        // to ensure there's nothing more
        var modifier = {
            type: act.type
        };
        updateAndCheckFields(actFieldsForType[act.type], act, modifier, null);

        // special check: customId must be unique for the user
        // TODO: if ownerUserID is null, will the unique constraint apply or not???
        if( Acts.findOne({ownerUserId: self.userId, customId: modifier.customId}) )
            throw new Meteor.Error("act-customId-not-unique");

        modifier.ownerUserId = self.userId;
        modifier.createdAt = new Date;
        modifier.modifiedAt = modifier.createdAt;
        modifier.executedAt = null;

        modifier._id = Acts.insert(modifier);
        return modifier;
    },

    // Updates existing act in database
    // Note that if one of the act fields is not defined (or undefined), it's not updated,
    // while if it's anything else including null, it will be updated!
    'updateAct': function(act) {
        var self = this;

        if( self.userId !== null )
            check(self.userId, String);

        check(act, Match.ObjectIncluding({
            _id: String
        }));

        var existingAct = Acts.findOne(act._id);

        // checks
        // NOTE: if the act has no owner, anybody who knows its _id can modify it (necessary to allow anonymous trials)
        if( !existingAct )
            throw new Meteor.Error("record-not-found");
        if( existingAct.ownerUserId && existingAct.ownerUserId !== self.userId )
            throw new Meteor.Error("insufficient-permissions"); // act owned by some other user
        if( !actFieldsForType[existingAct.type] || (act.type && act.type !== existingAct.type) )
            throw new Meteor.Error("act-type-invalid"); // act type cannot be changed

        // only update modify-able fields that are defined in 'act' and that are used by the actType
        var modifier = {};
        var update = false;
        if( act.ownerUserId !== self.userId ) {
            existingAct.ownerUserId = modifier.ownerUserId = self.userId; // associate to user after they log in
            update = true;
        }

        // if changing customId, ensure it's unique for the user
        // (do this before updateAndCheckFields because we need existingAct.customId!)
        // TODO: if ownerUserID is null, will the unique constraint apply or not???
        if( typeof(act.customId) === 'string' && act.customId !== existingAct.customId )
            if( Acts.findOne({ownerUserId: self.userId, customId: act.customId, _id: {$ne: act._id}}) )
                throw new Meteor.Error("act-customId-not-unique");

        // only update and check fields that actually changed value
        if( updateAndCheckFields(actFieldsForType[existingAct.type], act, modifier, existingAct) )
            update = true;

        // if anything changed, update the record
        if( update ) {
            existingAct.modifiedAt = modifier.modifiedAt = new Date;
            Acts.update(act._id, {$set: modifier});
        }

        return existingAct;
    },

    // Updates or inserts an act in database
    'upsertAct': function(act) {
        if( !act._id ) {
            delete act._id; // 'undefined' or 'null' is not enough, the field must be missing!
            return Meteor.call('insertAct', act);
        }  else
            return Meteor.call('updateAct', act);
    },

    'removeAct': function(actId) {
        var self = this;
        check(self.userId, String);
        check(actId, String);
        var act = Acts.findOne(actId);
        if (!act) {
            throw new Meteor.Error("record-not-found");
        }
        if (act.ownerUserId !== self.userId) {
            throw new Meteor.Error("insufficient-permissions");
        }
        // TODO: we'll have to periodically delete orphaned acts without owners, maybe an auto-delete index would do
        return Acts.remove(actId);
    },

    // TODO: move this to another file?
    'executeAct': function(actId, input) {
        var self = this;
        var userId = self.userId;

        check(actId, String);
        if( userId !== null )
            check(actId, String);

        // only run this on server
        if( self.isSimulation ) {
            return;
        }

        // Let other method calls from the same client start running,
        // without waiting for the HTTP request to complete.
        self.unblock();

        var act = Acts.findOne(actId);
        if (!act)
            throw new Meteor.Error("record-not-found", "Requested act was not found.");
        if( act.ownerUserId && act.ownerUserId !== self.userId )
            throw new Meteor.Error("insufficient-permissions", "Only owner can execute an act.");


        // invoke controller server
        var result = HTTP.post("http://localhost:1234/executeAct/" + actId, {
            timeout: 30000,
            data: {
                input: input || {}
            }
        });

        return result.data ? result.data.actExecutionId : null;

        /*var responsePattern = Match.ObjectIncluding({
            executionId: String,
            statusUrl: String,
            statusWsUrl: String});
        if( !Match.test(result.data, responsePattern)) {
            throw new Meteor.Error("unexpected-server-response", "The internal server returned an unexpected response.");
        }*/

        /*var actExecution = {};
        actExecution.actId = act._id;
        actExecution.startDate = new Date;
        actExecution.endDate = null;
        actExecution.script = act.script; // save a snapshot of the script, it might be changed later
        actExecution._id = ActExecutions.insert(actExecution);


        // open web socket to controller server
        var WebSocket = Meteor.npmRequire('ws');
        var ws = new WebSocket("ws://localhost:1234/executeAct");

        ws.on('open', Meteor.bindEnvironment( function() {
            var messageText = JSON.stringify({
                script: act.script
            });
            ws.send(messageText);
        }));


        ws.on('message', Meteor.bindEnvironment( function (messageText) {
            console.log('MSG received: %s', messageText);
            var msg = JSON.parse(messageText);
            ActExecutions.update(actExecution._id, {
                $push: { checkpoints: msg }
            });
        }));

        return actExecution._id; */

/*
        // invoke controller server
        var result = HTTP.post("http://localhost:1234/executeAct", {
            timeout: 30000,
            data: {
                script: act.script
            }
        });
        var responsePattern = Match.ObjectIncluding({
            executionId: String,
            statusUrl: String,
            statusWsUrl: String});
        if( !Match.test(result.data, responsePattern)) {
            throw new Meteor.Error("unexpected-server-response", "The internal server returned an unexpected response.");
        }
*/

    },

    'getCrawlFields': function() {
        return crawlFields;
    }
});
