/*

 This file contains a configuration describing user-editable fields of particular types of acts

 Act is an object like this:

 {
   _id: '<MongoDB ID>',
   createdAt: Date,
   modifiedAt: Date,
   executedAt: Date,
   type: 'crawl', // possible values 'crawl', 'script', ...

   // common fields
   customId: '...',
   apiActive: true

   // fields specific for act types
   startUrls: {...}

   // if any field is undefined or null, it's equally considered as not filled in.
   // for updates though, undefined means the field won't be updated
   // so you better use nulls for empty values
 }

*/


// database initialization
if (Meteor.isServer) {
    Meteor.startup(function () {
        // define indexes
        // (see http://joshowens.me/how-to-optimize-your-mongo-database-for-meteor-js/)
        Acts._ensureIndex({'ownerUserId': 1}, {sparse: 1});
        Acts._ensureIndex({'ownerUserId': 1, 'customId': 1}, {unique: 1, sparse: 1});

        // TODO: define initial data
        //if (Rooms.find().count() === 0) {
        //            Rooms.insert({name: "Initial room"});
        //        }
    });
}


// specification of user-editable act fields shared between all act types,
// will be copied into respective actFields
var commonActFields = {
    customId: {
        caption: 'ID',
        type: 'STRING',
        isMandatory: true,
        maxLength: 50,
        checkMessage: "ID must be a non-empty string"
    },

    apiActive: {
        caption: 'Act available through API',
        type: 'BOOLEAN'
    }
};


// specification of user-editable act fields
actFieldsForType = {

    // fields specific for crawl
    // TODO: is caption needed here? probably not...
    crawl : {
        startUrls: {
            caption: 'Start URLs',
            type: 'STRING_KEY_VALUE_ARRAY', // this must be [{key:'', val:''},...] to preserve order
            maxKeyLength: 50,
            maxValueLength: 1000,
            isKeyMandatory: false,
            isValueMandatory: true,
            checkMessage: "All start URLs must have an URL defined"
        },
        clickableElementsSelector: {
            caption: 'Clickable elements',
            type: 'STRING',
            maxLength: 1000
        },
        crawlPurls: {
            caption: 'Crawl PURLs',
            type: 'STRING_KEY_VALUE_ARRAY', // this must be [{key:'', val:''},...] to preserve order
            maxKeyLength: 50,
            maxValueLength: 1000,
            isKeyMandatory: false,
            isValueMandatory: true,
            checkMessage: "All crawl PURLs must have a pseudo-URL defined"
        },
        considerUrlFragment: {
            caption: 'Consider URL fragment',
            type: 'BOOLEAN'
        },
        maxCrawlDepth: {
            caption: 'Max crawling depth',
            type: 'INTEGER',
            min: 0,
            max: 999999999
        },
        maxCrawledPages: {
            caption: 'Max pages per crawl',
            type: 'INTEGER',
            min: 1,
            max: 999999999
        },
        maxCrawledPagesPerSlave: {
            caption: 'Max pages per slave',
            type: 'INTEGER',
            min: 1,
            max: 999999999
        },
        maxOutputPages: {
            caption: 'Max output pages',
            type: 'INTEGER',
            min: 1,
            max: 999999999
        },
        maxParallelRequests: {
            caption: 'Max parallel requests',
            type: 'INTEGER',
            min: 1,
            max: 999999999
        },
        loadImages: {
            caption: 'Load images',
            type: 'BOOLEAN',
            prefillValue: true
        },
        loadCss: {
            caption: 'Load CSS files',
            type: 'BOOLEAN',
            prefillValue: true
        },
        httpHeaders: {
            caption: 'Custom HTTP headers',
            type: 'STRING_KEY_VALUE_ARRAY',
            maxKeyLength: 50,
            maxValueLength: 1000,
            isKeyMandatory: true,
            isValueMandatory: false,
            checkMessage: "Custom HTTP headers cannot have an empty header name"
        },
        resourceTimeout: {
            caption: 'Resource timeout',
            // TODO: append "milliseconds" to end
            type: 'INTEGER',
            min: 100,
            max: 1000000
        },
        maxInfiniteScrollHeight: {
            caption: 'Max inifite scroll height',
            type: 'INTEGER',
            min: 0,
            max: 1000000
        },
        //{ name: 'randomWaitBetweenRequests', type: 'STRING' },
        injectJQuery: {
            caption: 'Inject jQuery',
            type: 'BOOLEAN',
            prefillValue: true
        },
        injectUnderscoreJs: {
            caption: 'Inject Underscore.js',
            type: 'BOOLEAN'
        },
        injectClientScripts: {
            caption: 'Inject client scripts',
            type: 'STRING' // TODO...
        },
        pageFunction: {
            caption: 'Page function',
            type: 'JAVASCRIPT_CODE',
            maxLength: 100000,
            prefillValue: "function pageFunction(context) {\n" +
                "  // called on every page the crawler visits, use it to extract data from it\n" +
                "  var $ = context.jQuery;\n" +
                "  var result = {\n" +
                "    myValue: $('TODO').text()\n" +
                "  };\n" +
                "  return result;\n" +
                "}"
        },
        interceptRequest: {
            caption: 'Intercept request function',
            type: 'JAVASCRIPT_CODE',
            maxLength: 100000,
            prefillValue: "function interceptRequest(context, newRequest) {\n" +
                "  // called whenever the crawler finds a link to a new page,\n" +
                "  // use it to override default behavior\n" +
                "  return newRequest;\n" +
                "}"
        }
    },

    // for demo only,
    /*script: {
        script: {
            caption: 'Script',
            type: 'JAVASCRIPT_CODE',
            maxLength: 100000,
            prefillValue: ""
        }
    }*/
};


// generate computed values for actFieldsForType
_.each( actFieldsForType, function( fields, actType ) {

    // extend fields with commonActFields
    _.each( commonActFields, function( commonField, commonFieldName ) {
        if( fields[commonFieldName] )
            throw 'Da foooook: Some type-specific act field has same name as common act field!!!';
        fields[commonFieldName] = commonField;
    });

    // extend each field with a Meteor-style checkPattern that describes
    // what field values can be stored in the database
    _.each( fields, function( fieldSpec, fieldName ) {

        // helper function
        var checkString = function(str, isMandatory, maxLength) {
            if( (str===null || str===undefined) && !isMandatory )
                return true;
            if( typeof(str)!=='string' )
                return false;
            if( isMandatory && str.length===0 )
                return false;
            return str.length <= maxLength;
        };

        switch( fieldSpec.type ) {
            case 'STRING':
            case 'JAVASCRIPT_CODE':
                fieldSpec.checkPattern = Match.Where( function (val) {
                    return checkString(val, fieldSpec.isMandatory, fieldSpec.maxLength);
                });
                break;

            case 'BOOLEAN':
                fieldSpec.checkPattern = Match.Where( function (val) {
                    if( (val===null || val===undefined) && !fieldSpec.isMandatory )
                        return true;
                    return typeof(val)==='boolean';
                });
                break;

            case 'INTEGER':
                fieldSpec.checkPattern = Match.Where( function (val) {
                    if( (val===null || val===undefined) && !fieldSpec.isMandatory )
                        return true;
                    return Match.test(val, Match.Integer) && fieldSpec.min <= val && val <= fieldSpec.max;
                });
                break;

            case 'STRING_KEY_VALUE_ARRAY':
                fieldSpec.checkPattern = Match.Where( function (array) {
                    if( (array === null || array === undefined) && !fieldSpec.isMandatory )
                        return true;
                    if( !Match.test(array, [{
                        key: Match.Where(function (key) {
                            return checkString(key, fieldSpec.isKeyMandatory, fieldSpec.maxKeyLength);
                        }),
                        value: Match.Where(function (value) {
                            return checkString(value, fieldSpec.isValueMandatory, fieldSpec.maxValueLength);
                        })}]))
                        return false;
                    return array.length > 0 || !fieldSpec.isMandatory;
                });
                break;

            default:
                throw "Da foooook: Unexpected act field type ("+fieldSpec.type+")!";
        }
    });

});





